import asyncio

from threading import Thread
from queue import Queue, Empty

from concurrent.futures._base import Executor
from functools import partial, wraps


class ThreadPoolExecutor(Executor):
    __slots__ = '__loop', '__futures', '__running', '__pool', '__tasks'

    def __init__(self, processes=None, loop: asyncio.AbstractEventLoop=None):
        self.__loop = loop or asyncio.get_event_loop()
        self.__futures = set()
        self.__running = True

        self.__pool = set()
        self.__tasks = Queue(0)

        for idx in range(processes):
            thread = Thread(
                target=self._in_thread,
                name="[%d] Thread Pool" % idx,
                daemon=True,
            )

            self.__pool.add(thread)
            # Starting the thread only after thread-pool will be started
            self.__loop.call_soon(thread.start)

        self.__pool = frozenset(self.__pool)

    def _in_thread(self):
        while self.__running:
            try:
                func, future = self.__tasks.get(timeout=1, block=True)
            except Empty:
                continue

            try:
                self.__loop.call_soon_threadsafe(future.set_result, func())
            except Exception as e:
                self.__loop.call_soon_threadsafe(future.set_exception, e)

    def submit(self, fn, *args, **kwargs):
        future = self.__loop.create_future()     # type: asyncio.Future
        self.__futures.add(future)
        future.add_done_callback(self.__futures.remove)
        self.__tasks.put_nowait((partial(fn, *args, **kwargs), future))
        return future

    def shutdown(self, wait=True):
        self.__running = False

        for f in filter(lambda x: not x.done(), self.__futures):
            f.set_exception(RuntimeError("Pool closed"))

    def __del__(self):
        self.shutdown()


def threaded(func):
    @wraps(func)
    async def wrap(*args, **kwargs):
        loop = asyncio.get_event_loop()

        return await loop.run_in_executor(
            None, partial(func, *args, **kwargs)
        )

    return wrap



import asyncio
import logging
import logging.handlers

from typing import Union

from aiomisc.thread_pool import threaded
from aiomisc.periodic import PeriodicCallback
from prettylog import (
    color_formatter,
    create_logging_handler,
    json_formatter,
    JSONLogFormatter,
    LogFormat,
    wrap_logging_handler as _wrap_logging_handler,
)


class AsyncMemoryHandler(logging.handlers.MemoryHandler):
    flush_async = threaded(logging.handlers.MemoryHandler.flush)


def wrap_logging_handler(handler: logging.Handler,
                         loop: asyncio.AbstractEventLoop=None,
                         buffer_size: int = 1024,
                         flush_interval: float = 0.1):
    loop = loop or asyncio.get_event_loop()

    buffered_handler = _wrap_logging_handler(
        handler=handler, buffer_size=buffer_size,
        logger_class=AsyncMemoryHandler,
    )

    periodic = PeriodicCallback(buffered_handler.flush_async)
    loop.call_soon_threadsafe(periodic.start, flush_interval, loop)

    return buffered_handler


def basic_config(level: int=logging.INFO,
                 log_format: Union[str, LogFormat]=LogFormat.color,
                 buffered=True, buffer_size: int=1024,
                 flush_interval: float=0.2, loop=None):

    if isinstance(level, str):
        level = getattr(logging, level.upper())

    logging.basicConfig()
    logger = logging.getLogger()
    logger.handlers.clear()

    if isinstance(log_format, str):
        log_format = LogFormat[log_format]

    handler = create_logging_handler(log_format)

    if buffered:
        handler = wrap_logging_handler(
            handler,
            buffer_size=buffer_size,
            flush_interval=flush_interval,
            loop=loop,
        )

    logging.basicConfig(
        level=level,
        handlers=[handler]
    )


__all__ = (
    'basic_config',
    'wrap_logging_handler',
    'JSONLogFormatter',
    'json_formatter',
    'color_formatter',
)



import asyncio
import itertools
import logging.handlers
import socket
from multiprocessing import cpu_count
from typing import Iterable, Any, List, Tuple

import uvloop

from aiomisc.thread_pool import ThreadPoolExecutor


log = logging.getLogger(__name__)


def chunk_list(iterable: Iterable[Any], size: int):
    iterable = iter(iterable)

    item = list(itertools.islice(iterable, size))
    while item:
        yield item
        item = list(itertools.islice(iterable, size))


OptionsType = List[Tuple[int, int, int]]


def bind_socket(*, address: str, port: int, options=()):
    if ':' in address:
        sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
    else:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    sock.setblocking(0)

    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)

    sock_addr = address, port

    if sock.family == socket.AF_INET6:
        log.info('Listening tcp://[%s]:%s' % sock_addr)
    else:
        log.info('Listening tcp://%s:%s' % sock_addr)

    try:
        sock.bind(sock_addr)
    finally:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 0)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 0)

    for level, option, value in options:
        sock.setsockopt(level, option, value)

    return sock


def new_event_loop(pool_size=None) -> asyncio.AbstractEventLoop:
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    pool_size = pool_size or cpu_count()

    asyncio.get_event_loop().close()

    loop = asyncio.new_event_loop()
    thread_pool = ThreadPoolExecutor(pool_size, loop=loop)

    loop.set_default_executor(thread_pool)

    asyncio.set_event_loop(loop)

    return loop
